#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/eeprom.h>

/*
Telefonumbau für Alpha Ferro und RFT Variant auf Tonwahl
ATtiny24  = 4 MHz
FUSE-HIGH = 0xDD
FUSE-LOW  = 0xDC

Tonausgabe über Timer1 OC1A=PA6
Nummernschalter und Erdtase = Eingänge

Erdtaste fungiert als Shift-Taste mit der die Zweitbelegung gewählt wird...
Taste drücken, Nummernschalter aufziehen, loslassen, Erdtaste loslassen
Hier jetzt die Zweitbelegung definieren ggf. ändern --> "1234567890ABCD*#X" <--- X=Wahlwiederholung
*/
static const __flash uint8_t ShiftTable[10] = {
  'X', // ET&0 = Wahlwiederholung
  '*', // ET&1 = *
  '#', // ET&2 = #
  'A', // ET&3 = A
  'B', // ET&4 = B
  'C', // ET&5 = C
  'D', // ET&6 = D
  'X', // ET&7 = Wahlwiederholung
  'X', // ET&8 = Wahlwiederholung
  'X', // ET&9 = Wahlwiederholung
};

// Dauer Ton-, Pausenzeiten ggf. ändern
#define ON_TIME_MS  250
#define OFF_TIME_MS 100

// Ein-, Ausgänge
#define IO_ET       PA3                           // Erdtaste -> Low wenn gedrückt
#define IO_NSI      PA4                           // NSI Low wenn IDLE
#define IO_NSA      PA5                           // NSA Low wenn Aktivität am Nummernschalter
#define IO_DTMF     PA6                           // Tonausgang

// Nummernspeicher für Wahlwiederholung im EEPROM in ASCII Form (in 48 Bytes kann man Lieder speichern)
#define EEBYTES 48
uint8_t EEMEM emem[EEBYTES];

// Tonfrequenzvariablen, im main füllen, werden im Tim1OverflowIRQ gelesen
static volatile uint8_t X_swa=0;
static volatile uint8_t X_swb=0;




/***********************************************************************************************************************
Timer1 Funktionen - Tonausgabe
***********************************************************************************************************************/
ISR(TIM1_OVF_vect)
{
  static const __flash uint8_t sin_table[128] = {
    63, 66, 69, 72, 75, 79, 82, 85, 87, 90, 93, 96, 99, 101, 104, 106,
    108, 110, 112, 114, 116, 118, 119, 121, 122, 123, 124, 125, 126, 126, 127, 127,
    127, 127, 127, 126, 126, 125, 124, 123, 122, 121, 119, 118, 116, 114, 112, 110,
    108, 106, 104, 101, 99, 96, 93, 90, 87, 85, 82, 79, 75, 72, 69, 66,
    63, 60, 57, 54, 51, 47, 44, 41, 39, 36, 33, 30, 27, 25, 22, 20,
    18, 16, 14, 12, 10, 8, 7, 5, 4, 3, 2, 1, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 1, 2, 3, 4, 5, 7, 8, 10, 12, 14, 16,
    18, 20, 22, 25, 27, 30, 33, 36, 39, 41, 44, 47, 51, 54, 57, 60
  };
  static uint16_t cursina = 0;                    // position freq. A in LUT (extended format)
  static uint16_t cursinb = 0;                    // position freq. B in LUT (extended format)
  uint16_t tmpsina;                               // position freq. A in LUT (actual position)
  uint16_t tmpsinb;                               // position freq. B in LUT (actual position)

  // move Pointer about step width aheaed
  cursina += X_swa;
  cursinb += X_swb;
  // normalize Temp-Pointer
  tmpsina  =  (uint8_t)(((cursina+4) >> 3)&(0x007F));
  tmpsinb  =  (uint8_t)(((cursinb+4) >> 3)&(0x007F));
  // calculate PWM value: high frequency value + 3/4 low frequency value
  OCR1A = (sin_table[tmpsina] + (sin_table[tmpsinb]-(sin_table[tmpsinb]>>2)));
}
// ---------------------------------------------------------------------------------------------------------------------
// Tonausgabe stoppen
static void stop_dtmf(void)
{
  X_swb=0;
  X_swb=0;
  TIMSK1 &=0b11111110;                            // Timer1 Overflow-IRQ aus
  TCCR1A &=0b01111111;                            // Timer von Ausgang trennen
  PORTA &= ~(1<<IO_DTMF);                         // Ausgang manuell
  DDRA &= ~(1<<IO_DTMF);                          // auf High-Z setzen
 }
// ---------------------------------------------------------------------------------------------------------------------
// Tonausgabe starten
static void start_dtmf(void)
{
  DDRA |= (1<<IO_DTMF);                           // als Ausgang setzen
  TCCR1A |=0b10000001;                            // Clear OC1A on Compare Match, Timer1 Mode1
  TCCR1B |=0b00000001;                            // Timer1 Prescaler = /1
  TIMSK1 |=0b00000001;                            // Timer1 Overflow-IRQ ein
}
// ---------------------------------------------------------------------------------------------------------------------
// Einzelton ausgeben (200ms blockierend) -> Übergabe = Ziffer in ASCII Format "0123456789ABCD*#!"
//
//       1209Hz 1336Hz 1477Hz 1633Hz
// 697Hz   1      2      3      A
// 770Hz   4      5      6      B
// 852Hz   7      8      9      C
// 941Hz   *      0      #      D
//
// ! als Quittierungston
static void send_dtmf(uint8_t ascii_val)
{
  const uint8_t dtmf_freqL[4] = {91, 101, 111, 123};
  const uint8_t dtmf_freqH[4] = {158, 174, 193, 213};

  switch (ascii_val) {
    case '!': X_swb = 50; X_swa = 250; break;

    case '1': X_swb = dtmf_freqL[0]; X_swa = dtmf_freqH[0]; break;
    case '2': X_swb = dtmf_freqL[0]; X_swa = dtmf_freqH[1]; break;
    case '3': X_swb = dtmf_freqL[0]; X_swa = dtmf_freqH[2]; break;
    case 'A': X_swb = dtmf_freqL[0]; X_swa = dtmf_freqH[3]; break;

    case '4': X_swb = dtmf_freqL[1]; X_swa = dtmf_freqH[0]; break;
    case '5': X_swb = dtmf_freqL[1]; X_swa = dtmf_freqH[1]; break;
    case '6': X_swb = dtmf_freqL[1]; X_swa = dtmf_freqH[2]; break;
    case 'B': X_swb = dtmf_freqL[1]; X_swa = dtmf_freqH[3]; break;

    case '7': X_swb = dtmf_freqL[2]; X_swa = dtmf_freqH[0]; break;
    case '8': X_swb = dtmf_freqL[2]; X_swa = dtmf_freqH[1]; break;
    case '9': X_swb = dtmf_freqL[2]; X_swa = dtmf_freqH[2]; break;
    case 'C': X_swb = dtmf_freqL[2]; X_swa = dtmf_freqH[3]; break;

    case '*': X_swb = dtmf_freqL[3]; X_swa = dtmf_freqH[0]; break;
    case '0': X_swb = dtmf_freqL[3]; X_swa = dtmf_freqH[1]; break;
    case '#': X_swb = dtmf_freqL[3]; X_swa = dtmf_freqH[2]; break;
    case 'D': X_swb = dtmf_freqL[3]; X_swa = dtmf_freqH[3]; break;

    default: return;
  }

  start_dtmf();
  _delay_ms(ON_TIME_MS);
  stop_dtmf();
}
/***********************************************************************************************************************
EEPROM Funktionen -> Nummernspeicher
***********************************************************************************************************************/
// Fügt übergebenes Zeichen dem EEPROM Wahlwiederholungsspeicher zu
static void eedigit_store(uint8_t ascii_val) {
  static uint8_t ep=0;                            // Bei jedem mal Hörer abnehmen/Controllerstart

  if (ep >= (EEBYTES-1)) return;                  // Nur begrenzt Speicher
  eeprom_write_byte (&emem[ep], ascii_val);       // Ziffer speichern/zufügen
  ep++;
  eeprom_write_byte (&emem[ep], '\0');            // Nullbyte anfügen (Ende-Kennzeichen)
}
// ---------------------------------------------------------------------------------------------------------------------
// Liest aktuelle Rufnummer aus Wahlwiederholungsspeicher und gibt diese als Tonfolge aus
static void eemem_send(void) {
  uint8_t val;
  uint8_t ep;

  for (ep=0; ep<EEBYTES; ep++) {                  // Nummer ist als String
    val = eeprom_read_byte (&emem[ep]);           // im ASCII-Format im EEPROM gespeichert
    if (val == 0) break;                          // auslesen bis 0-Byte
    else {
      send_dtmf(val);                             // Tonausgabe
      _delay_ms(OFF_TIME_MS);                     // Pause zwischen den Tönen
    }
  }
}
/***********************************************************************************************************************
Hauptfunktionen
***********************************************************************************************************************/
// ---------------------------------------------------------------------------------------------------------------------
// Pulse des Nummernschalters werden gezählt und diese werden in ASCII Form zurückgegeben: "1234567890ABCD*#X"
// Bei Fehlerkennung wird 0 --> Nullbyte zurückgegeben
// Routine anspringen bei High-Low Flanke nsa --> Start des Bereichs B unten
//
// Signalverlauf, Timing an Eingängen nsi, nsa
// A=Nummernschalter ruht ; B=Nummernschalter wird aufgezogen ; C=Nummernschalter dreht zurück
//
//    <--A---><--------B--------><------------------C-----------------><----A-----><------B------><-----C-----><---A----
//    -------                                                           ----------                              --------
// nsa       |                                                         |          |                            |
//            ---------------------------------------------------------            ----------------------------
//                               --   --   --   --   --   --   --   --                             --   --   --
// nsi                          |  | |  | |  | |  | |  | |  | |  | |  |                           |  | |  | |  |
//    --------------------------    -    -    -    -    -    -    -    ---------------------------    -    -    --------
//
// High/Low-Dauer=60ms Puls,40ms Pause <- 100ms
// gwählt wurde hier 8 3
static uint8_t read_nsi(void)
{
  uint8_t shift= PINA &(1<<IO_ET);                // Shift-Taste bei Funktionseintritt
  uint8_t last_nsi= PINA &(1<<IO_NSI);
  uint8_t cur_nsi= PINA &(1<<IO_NSI);
  uint8_t cntms=0;
  uint8_t ret=0;

  while ((PINA &(1<<IO_NSA)) == 0) {              // Ende wenn NSA auf High geht
    _delay_ms(1);                                 // nsi im Abstand von 1ms samplen
    cntms++;
    last_nsi=cur_nsi;
    cur_nsi= PINA &(1<<IO_NSI);
    if ((last_nsi == 0) && (cur_nsi != 0)) {      // Steigende Flanke
      cntms=0;                                    // Counter reset
    }
    else if ((last_nsi != 0) && (cur_nsi == 0)) { // Fallende Flanke
      if ((cntms >= 40) && (cntms <= 80)) {       // 60ms High Puls +-20ms
        ret++;                                    // Pulszähler (1...10)
      }
    }
  }

  if ((ret == 0) || (ret > 10)) return 0;         // Fehlerkennung --> kein Ton
  if (ret == 10) ret = 0;                         // 10 Pulse stehen für Ziffer 0
  if (shift != 0) ret +='0';                      // OHNE Erdtaste 0...9 nach ASCII konvertieren
  else ret=ShiftTable[ret];                       // MIT Shift Taste (globale Zweitbelegung)
  return ret;
}
// ---------------------------------------------------------------------------------------------------------------------
// Hauptschleife wenn Höher mit ET-Taste abgehoben wird --> Funktionstest
// Je nach gewählter Ziffer werden 1...10 Piepser ausgegeben
static void mainloop_debug(void) {
  uint8_t nsival;
  uint8_t cnt;

  for (cnt=5; cnt>0; cnt--) {                     // Startquittierung
    send_dtmf('!');
    _delay_ms(250);
  }

  while (1) {
    if ((PINA &(1<<IO_NSA)) == 0) {               // Wenn Nummernschalter aufgezogen wird
      nsival = read_nsi();                        // Ziffer lesen (ASCII-Format)
      if (nsival != 0) {                          // Wenn gültig ASCII --> Binär
        if (nsival == '0') nsival=10; else nsival = nsival - '0';
        while (nsival > 0) {                      // 1-> 1mal Piep ... 0-> 10mal Piep
          send_dtmf('!');                         // Quittierungston (KEIN DTMF)
          _delay_ms(250);
          nsival--;
        }
      }
    }
  }
}
// ---------------------------------------------------------------------------------------------------------------------
// Hauptschleife Normalfunktion wenn Höhrer OHNE ET Taste abgenommen wird
static void mainloop_normal(void) {
 uint8_t nsival;

  while (1) {
    if ((PINA &(1<<IO_NSA)) == 0) {               // Wenn Nummernschalter aufgezogen wird
      nsival = read_nsi();                        // Ziffer lesen (ASCII-Format)
      if (nsival != 0) {                          // Wenn keine Fehlerkennung
        if (nsival == 'X') eemem_send();          // Wahlwiederholung wurde gewählt <-- Senden ohne speichern
        else {                                    // normale Ziffer/Zeichen "1234567890ABCD*#"
          eedigit_store(nsival);                  // Ziffer zum Wahlwiederholungsspeicher zufügen
          send_dtmf(nsival);                      // Tonausgabe für Ziffer
        }
      }
    }
  }
}
// ---------------------------------------------------------------------------------------------------------------------
// Controllerstart (wenn Höhrer abgenommen wird) MIT gedrückter Erdtaste = Piepfunktion zur Kontrolle
// Ausgabe Anzahl Piepser je nach gewählter Ziffer
// Controllerstart OHNE gedrückte Erdtaste = Normalfunktion
// Jede Ziffer wird als DTMF Ton ausgegeben
int main (void)
{
  DDRA=  0b10000111;                              // PA7=NC, PA6=DTMF-Ausgang=High-Z, PA5-PA3=Eingänge, PA2-PA0=NC
  PORTA= 0b00111000;                              // PA5-PA3 interne Pull-Ups=Ein
  sei();

  uint8_t cur = PINA &(1<<IO_ET);
  _delay_ms(50);
  uint8_t now = PINA &(1<<IO_ET);
  if ((cur == 0) && (now == 0)) mainloop_debug();
  else mainloop_normal();
}
// ---------------------------------------------------------------------------------------------------------------------
